/*
 * ftpd/ftpd.c --
 *
 *	A very simple and basic ftp daemon. This file contains the
 *	mainloop and the command implementations.
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "ftpd.h"

static void cmd_cwd(FILE *client, int argc, char **argv);
static void cmd_epsv(FILE *client, int argc, char **argv);
static void cmd_help(FILE *client, int argc, char **argv);
static void cmd_list(FILE *client, int argc, char **argv);
static void cmd_nlst(FILE *client, int argc, char **argv);
static void cmd_mkd(FILE *client, int argc, char **argv);
static void cmd_pass(FILE *client, int argc, char **argv);
static void cmd_pasv(FILE *client, int argc, char **argv);
static void cmd_pwd(FILE *client, int argc, char **argv);
static void cmd_quit(FILE *client, int argc, char **argv);
static void cmd_rmd(FILE *client, int argc, char **argv);
static void cmd_stat(FILE *client, int argc, char **argv);
static void cmd_user(FILE *client, int argc, char **argv);

typedef enum {
    CONNECTED, USER, AUTHENTICATED
} state_t;

static state_t state;

const char *progname = "ftpd";

typedef struct cmd {
    char *name;
    void (*func) (FILE *client, int argc, char **argv);
    state_t state;
} cmd_t;

static cmd_t cmds[] = {
    { "CWD",  cmd_cwd,  AUTHENTICATED },
    { "EPSV", cmd_epsv, AUTHENTICATED },
    { "HELP", cmd_help, CONNECTED },
    { "LIST", cmd_list, AUTHENTICATED },
    { "NLST", cmd_nlst, AUTHENTICATED },
    { "MKD",  cmd_mkd,  AUTHENTICATED },
    { "PASS", cmd_pass, USER },
    { "PASV", cmd_pasv, AUTHENTICATED },
    { "PWD",  cmd_pwd,  AUTHENTICATED },
    { "QUIT", cmd_quit, CONNECTED },
    { "RMD",  cmd_rmd,  AUTHENTICATED },
    { "STAT", cmd_stat, AUTHENTICATED },
    { "USER", cmd_user, CONNECTED },
    { NULL, NULL }
};

static char *host = NULL;
static char *user = NULL;
static char *password = NULL;
static char *prefix = NULL;
static char *config = NULL;
static char *cwd = NULL;
static int pasv = -1;		/* passive socket for data transfers */

static char  *data_ports[] = {
    "54321", "43434", "54326", "12891", NULL
};

static void
cmd_cwd(FILE *client, int argc, char **argv)
{
    char buffer[MAXPATHLEN];
    char *newcwd;
    
    if (argc != 2) {
	fprintf(client, "501 wrong number of arguments\r\n");
	return;
    }

    snprintf(buffer, sizeof(buffer), "%s/%s", cwd, argv[1]);
    newcwd = normalize(buffer);
    if (! newcwd) {
	fprintf(client, "550 failed to change directory to %s\r\n", buffer);
	return;
    }
    
    snprintf(buffer, sizeof(buffer), "%s/%s", prefix, newcwd);
    if (chdir(buffer) < 0) {
	fprintf(client, "550 failed to change to directory %s\r\n", newcwd);
	free(newcwd);
	return;
    }

    if (cwd) free(cwd);
    cwd = newcwd;
    
    fprintf(client, "250 changed directory to %s\r\n", cwd);
}

static void
cmd_epsv(FILE *client, int argc, char **argv)
{
    struct sockaddr_storage saddr;
    socklen_t saddrlen = sizeof(saddr);
    char port[NI_MAXSERV];
    int i, n;
    
    if (argc != 1) {
        fprintf(client, "501 too many arguments\r\n");
	return;
    }

    if (pasv > -1) {
	(void) close(pasv);
	pasv = -1;
    }

    for (i = 0; pasv == -1 && data_ports[i]; i++) {
	pasv = tcp_listen(data_ports[i]);
    }
    if (pasv == -1) {
	fprintf(client, "425 can't open listening port\r\n");
	return;
    }

    n = getsockname(pasv, (struct sockaddr *) &saddr, &saddrlen);
    if (n == -1) {
	fprintf(client, "425 can't get socket name\r\n");
	(void) close(pasv);
	pasv = -1;
	return;
    }

    n = getnameinfo((struct sockaddr *) &saddr, saddrlen,
		    NULL, 0, port, sizeof(port), NI_NUMERICSERV);
    if (n) {
	fprintf(client, "425 can't convert port number\r\n");
	(void) close(pasv);
	pasv = -1;
	return;
    }

    fprintf(client, "227 Entering Extended Passive Mode (|||%s|)\r\n", port);
}

static void
cmd_help(FILE *client, int argc, char **argv)
{
    int i;
    const int columns = 6;

    if (argc != 1) {
	fprintf(client, "501 too many arguments\r\n");
	return;
    }

    fprintf(client, "214-The following commands are recognized:");
    for (i = 0; cmds[i].name; i++) {
	if ((i % columns) == 0) {
	    fprintf(client, "\r\n ");
	}
	fprintf(client, "%s\t", cmds[i].name);
    }
    if (! (i % columns) == 0) fprintf(client, "\r\n");
    fprintf(client, "214 Direct comments to "
	    "<j.schoenwaelder@jacobs-university.de>\r\n");
}

static void
cmd_list(FILE *client, int argc, char **argv)
{
    char buffer[MAXPATHLEN];
    int fd;
    FILE *stream;

    if (argc > 2) {
        fprintf(client, "501 wrong number of arguments\r\n");
    }

    if (argc == 2) {
	snprintf(buffer, sizeof(buffer), "%s/%s/%s", prefix, cwd, argv[1]);
    } else {
	snprintf(buffer, sizeof(buffer), "%s/%s", prefix, cwd);
    }

    if (pasv == -1) {
	fprintf(client, "425 Use PASV or EPSV first\r\n");
	return;
    }

    fprintf(client, "150 File status okay; about to open data connection\r\n");
    fd = tcp_accept(pasv);
    if (fd == -1) {
	fprintf(client, "451 data transfer aborted; accept failed\r\n");
	return;
    }

    stream = fdopen(fd, "w");
    if (! stream) {
	fprintf(client, "451 data transfer aborted; opening stream failed\r\n");
	close(fd);
	return;
    }

    ls(stream, buffer, FTPD_LS_LONG);
    if (fflush(stream) || ferror(stream)) {
	fprintf(client, "451 data transfer failed; stream I/O error\r\n");
	close(fd);
	return;
    }

    fprintf(client, "226 Closing data connection\r\n");
    fclose(stream);
}

static void
cmd_nlst(FILE *client, int argc, char **argv)
{
    char buffer[MAXPATHLEN];
    int fd;
    FILE *stream;

    if (argc > 2) {
        fprintf(client, "501 wrong number of arguments\r\n");
    }

    if (argc == 2) {
	snprintf(buffer, sizeof(buffer), "%s/%s/%s", prefix, cwd, argv[1]);
    } else {
	snprintf(buffer, sizeof(buffer), "%s/%s", prefix, cwd);
    }

    if (pasv == -1) {
	fprintf(client, "425 Use PASV or EPSV first\r\n");
	return;
    }

    fprintf(client, "150 File status okay; about to open data connection\r\n");
    fd = tcp_accept(pasv);
    if (fd == -1) {
	fprintf(client, "451 data transfer aborted; accept failed\r\n");
	return;
    }

    stream = fdopen(fd, "rw");
    if (! stream) {
	fprintf(client, "451 data transfer aborted; opening stream failed\r\n");
	close(fd);
	return;
    }

    ls(stream, buffer, FTPD_LS_NONE);
    if (fflush(stream) || ferror(stream)) {
	fprintf(client, "451 data transfer failed; stream I/O error\r\n");
	close(fd);
	return;
    }

    fprintf(client, "226 Closing data connection\r\n");
    fclose(stream);
}

static void
cmd_mkd(FILE *client, int argc, char **argv)
{
    char buffer[MAXPATHLEN];
    char *path;
    
    if (argc != 2) {
	fprintf(client, "501 wrong number of arguments\r\n");
	return;
    }

    snprintf(buffer, sizeof(buffer), "%s/%s", cwd, argv[1]);
    path = normalize(buffer);
    if (! path) {
	fprintf(client, "550 failed to make directory %s/%s\r\n", cwd, argv[1]);
	return;
    }
    
    snprintf(buffer, sizeof(buffer), "%s/%s", prefix, path);
    if (mkdir(buffer, 0777) < 0) {
	fprintf(client, "550 failed to make directory %s\r\n", path);
	free(path);
	return;
    }

    fprintf(client, "257 created directory %s\r\n", path);
    free(path);
}

static void
cmd_pass(FILE *client, int argc, char **argv)
{
    if (argc != 2) {
        fprintf(client, "501 wrong number of arguments\r\n");
	return;
    }

    if (! password || strcmp(argv[1], password) != 0) {
	fprintf(client, "530 Login incorrect.\r\n");
	return;
    }
    
    fprintf(client, "230 Login successful.\r\n");
    state = AUTHENTICATED;
    if (cwd) free(cwd);
    cwd = strdup("/");
}

static void
cmd_pasv(FILE *client, int argc, char **argv)
{
    struct sockaddr_storage saddr;
    socklen_t saddrlen = sizeof(saddr);
    struct sockaddr_in *saddr_in;
    int i, n;
    char buf[20];
    
    if (argc != 1) {
        fprintf(client, "501 too many arguments\r\n");
	return;
    }

    if (pasv > -1) {
	(void) close(pasv);
	pasv = -1;
    }

    for (i = 0; pasv == -1 && data_ports[i]; i++) {
	pasv = tcp_listen(data_ports[i]);
    }
    if (pasv == -1) {
        fprintf(client, "425 can't open listening port\r\n");
	return;
    }

    n = getsockname(pasv, (struct sockaddr *) &saddr, &saddrlen);
    if (n == -1) {
	fprintf(client, "425 can't get socket name\r\n");
	(void) close(pasv);
	pasv = -1;
	return;
    }

    if (saddr.ss_family != AF_INET) {
	fprintf(client, "425 can't convert port number; wrong address family\r\n");
	(void) close(pasv);
	pasv = -1;
	return;
    }

    saddr_in = (struct sockaddr_in *) &saddr;
    inet_ntop(saddr_in->sin_family, &saddr_in->sin_addr, buf, sizeof(buf));
    for (i = 0; buf[i]; i++) {
	if (buf[i] == '.') buf[i] = ',';
    }
    
    fprintf(client, "227 Entering Passive Mode (%s,%d,%d)\r\n", buf,
	    ntohs(saddr_in->sin_port) / 256, ntohs(saddr_in->sin_port) % 256);
}

static void
cmd_pwd(FILE *client, int argc, char **argv)
{
    if (argc != 1) {
	fprintf(client, "501 too many arguments\r\n");
	return;
    }

    fprintf(client, "257 \"%s\"\r\n", cwd);
}

static void
cmd_quit(FILE *client, int argc, char **argv)
{
    if (argc != 1) {
	fprintf(client, "501 too many arguments\r\n");
	return;
    }
    
    fprintf(client, "221 Goodbye.\r\n");
    fflush(client);
    fclose(client);
    exit(EXIT_SUCCESS);
}

static void
cmd_rmd(FILE *client, int argc, char **argv)
{
    char buffer[MAXPATHLEN];
    char *path;

    if (argc != 2) {
	fprintf(client, "501 wrong number of arguments\r\n");
	return;
    }

    snprintf(buffer, sizeof(buffer), "%s/%s", cwd, argv[1]);
    path = normalize(buffer);
    if (! path) {
	fprintf(client, "550 failed to remove directory %s/%s\r\n", cwd, argv[1]);
	return;
    }

    snprintf(buffer, sizeof(buffer), "%s/%s", prefix, path);
    if (rmdir(buffer) < 0) {
	fprintf(client, "550 failed to remove directory %s\r\n", path);
	free(path);
	return;
    }

    fprintf(client, "250 removed directory %s\r\n", path);
    free(path);
}

static void
cmd_stat(FILE *client, int argc, char **argv)
{
    char buffer[MAXPATHLEN];

    if (argc == 1) {
	fprintf(client, "211-FTP server status:\r\n");
	fprintf(client, "  Host: %s\r\n", host ? host : "<unknown>");
	fprintf(client, "  User: %s\r\n", user ? user : "<unknown>");
	fprintf(client, "  Path: %s\r\n", cwd ? cwd : "<unknown>");
	fprintf(client, "211 End of status\r\n");
    } else if (argc == 2) {
	snprintf(buffer, sizeof(buffer), "%s/%s/%s", prefix, cwd, argv[1]);
	fprintf(client, "213-Status follows:\r\n");
	ls(client, buffer, FTPD_LS_LONG);
	fprintf(client, "213 End of status\r\n");
    } else {
        fprintf(client, "501 wrong number of arguments\r\n");
    }
}

static void
cmd_user(FILE *client, int argc, char **argv)
{
    if (argc != 2) {
        fprintf(client, "501 wrong number of arguments\r\n");
	return;
    }

    state = CONNECTED;
    if (user) { free(user); user = NULL; }
    if (password) { free(password); password = NULL; }
    if (prefix) { free(prefix); prefix = NULL; }
    
    state = USER;
    user = strdup(argv[1]);
    (void) get_config(config, user, &password, &prefix);
    fprintf(client, "331 Please specify the password.\r\n");
}

static char*
next_token(char **s)
{
    char *token, *p;

    assert(s && *s);

    for (p = *s; *p && isspace(*p); p++) {
	*p = '\0';
    }
    token = p;
    for (; *p && !isspace(*p); p++) ;
    for (; *p && isspace(*p); p++) {
	*p = '\0';
    }
    *s = p;
    return token;
}

static void
read_command(FILE *client, int *argc, char ***args)
{
    static char line[BUFLEN];
    static char* argv[BUFLEN];
    char *p;

    assert(client && args);

    memset((char *) argv, 0, sizeof(argv));
    p = fgets(line, sizeof(line), client);
    for (*argc = 0; p && *p; (*argc)++) {
	argv[*argc] = next_token(&p);
    }
    *args = argv;
}

static int
mainloop(FILE *client)
{
    struct sockaddr_storage peer;
    socklen_t peerlen = sizeof(peer);
    char buffer[NI_MAXHOST];
    int i, n, argc;
    char **argv;

    assert(client);

    n = getpeername(fileno(client), (struct sockaddr *) &peer, &peerlen);
    if (n == 0) {
	n = getnameinfo((struct sockaddr *) &peer, peerlen,
			buffer, sizeof(buffer), NULL, 0,
			NI_NUMERICHOST);
	if (n == 0) {
	    host = strdup(buffer);
	}
    }

    state = CONNECTED;
    
    fprintf(client, "220 js-ftpd server ready\r\n");
    fflush(client);

    while (1) {
	read_command(client, &argc, &argv);
	if (argv[0] == NULL) {
	    fprintf(client, "421 service not available (parsing error)\r\n");
	    return EXIT_FAILURE;
	}
	if (strlen(argv[0]) == 0) {
	    continue;
	}

	for (i = 0; cmds[i].name; i++) {
	    if (strcasecmp(argv[0], cmds[i].name) == 0) break;
	}

	if (cmds[i].func) {
	    if (cmds[i].state > state) {
		fprintf(client, "530 Please login with USER and PASS.\r\n");
	    } else {
		(cmds[i].func) (client, argc, argv);
	    }
	} else {
	    fprintf(client, "500 Unknown command.\r\n");
	}
	if (fflush(client) || ferror(client)) {
	    syslog(LOG_ERR, "communication failure - exiting");
	    exit(EXIT_FAILURE);
	}
    }
    
    return EXIT_SUCCESS;
}

int
main(int argc, char **argv)
{
    int lfd, cfd;
    pid_t pid;
    
    if (argc != 2) {
        fprintf(stderr, "usage: %s port\n", progname);
        exit(EXIT_FAILURE);
    }

    config = strdup("./ftpd-user.conf");

    openlog(progname, LOG_PID, LOG_DAEMON);

    lfd = tcp_listen(argv[1]);

    while (lfd != -1) {
	cfd = tcp_accept(lfd);
	if (cfd == -1) {
	    continue;
	}

	pid = fork();
	if (pid == -1) {
	    syslog(LOG_ERR, "fork failed: %s", strerror(errno));
	    close(cfd);
	    continue;
	}

	/* Fork of a client process to handle this connection. We like
	 * to use stream I/O since this is somewhat simpler (we only
	 * have to remember to flush at appropriate times). */

	if (pid == 0) {
	    FILE *client;
	    client = fdopen(cfd, "w+");
	    if (! client) {
		syslog(LOG_ERR, "fdopen failed: %s", strerror(errno));
		close(cfd);
		continue;
	    }

	    exit(mainloop(client));
	} else {
	    close(cfd);
	}
    }

    tcp_close(lfd);

    closelog();

    return EXIT_SUCCESS;
}
